__all__ = ['EVENT_HANDLER_EVENT_TYPE_MAP']

EVENT_HANDLER_EVENT_TYPE_MAP: dict[str, str] = {
    'onabort': 'abort',
    'onauxclick': 'auxclick',
    'onbeforeinput': 'beforeinput',
    'onbeforematch': 'beforematch',
    'onbeforetoggle': 'beforetoggle',
    'oncancel': 'cancel',
    'oncanplay': 'canplay',
    'oncanplaythrough': 'canplaythrough',
    'onchange': 'change',
    'onclick': 'click',
    'onclose': 'close',
    'oncontextlost': 'contextlost',
    'oncontextmenu': 'contextmenu',
    'oncontextrestored': 'contextrestored',
    'oncopy': 'copy',
    'oncuechange': 'cuechange',
    'oncut': 'cut',
    'ondblclick': 'dblclick',
    'ondrag': 'drag',
    'ondragend': 'dragend',
    'ondragenter': 'dragenter',
    'ondragleave': 'dragleave',
    'ondragover': 'dragover',
    'ondragstart': 'dragstart',
    'ondrop': 'drop',
    'ondurationchange': 'durationchange',
    'onemptied': 'emptied',
    'onended': 'ended',
    'onformdata': 'formdata',
    'oninput': 'input',
    'oninvalid': 'invalid',
    'onkeydown': 'keydown',
    'onkeypress': 'keypress',
    'onkeyup': 'keyup',
    'onloadeddata': 'loadeddata',
    'onloadedmetadata': 'loadedmetadata',
    'onloadstart': 'loadstart',
    'onmousedown': 'mousedown',
    'onmouseenter': 'mouseenter',
    'onmouseleave': 'mouseleave',
    'onmousemove': 'mousemove',
    'onmouseout': 'mouseout',
    'onmouseover': 'mouseover',
    'onmouseup': 'mouseup',
    'onpaste': 'paste',
    'onpause': 'pause',
    'onplay': 'play',
    'onplaying': 'playing',
    'onprogress': 'progress',
    'onratechange': 'ratechange',
    'onreset': 'reset',
    'onsecuritypolicyviolation': 'securitypolicyviolation',
    'onseeked': 'seeked',
    'onseeking': 'seeking',
    'onselect': 'select',
    'onslotchange': 'slotchange',
    'onstalled': 'stalled',
    'onsubmit': 'submit',
    'onsuspend': 'suspend',
    'ontimeupdate': 'timeupdate',
    'ontoggle': 'toggle',
    'onvolumechange': 'volumechange',
    'onwaiting': 'waiting',
    'onwebkitanimationend': 'webkitAnimationEnd',
    'onwebkitanimationiteration': 'webkitAnimationIteration',
    'onwebkitanimationstart': 'webkitAnimationStart',
    'onwebkittransitionend': 'webkitTransitionEnd',
    'onwheel': 'wheel',
    'onblur': 'blur',
    'onerror': 'error',
    'onfocus': 'focus',
    'onload': 'load',
    'onresize': 'resize',
    'onscroll': 'scroll',
    'onscrollend': 'scrollend',
    'onafterprint': 'afterprint',
    'onbeforeprint': 'beforeprint',
    'onbeforeunload': 'beforeunload',
    'onhashchange': 'hashchange',
    'onlanguagechange': 'languagechange',
    'onmessage': 'message',
    'onmessageerror': 'messageerror',
    'onoffline': 'offline',
    'ononline': 'online',
    'onpagehide': 'pagehide',
    'onpageshow': 'pageshow',
    'onpopstate': 'popstate',
    'onrejectionhandled': 'rejectionhandled ',
    'onstorage': 'storage',
    'onunhandledrejection': 'unhandledrejection',
    'onunload': 'unload',
    'onreadystatechange': 'readystatechange',
    'onvisibilitychange': 'visibilitychange',

    # custom
    '_oncontentchange': '_contentchange',
}



__all__ = [
    'Event',
    'Gladius',
]

import os
import asyncio
import inspect
from uuid import uuid4
from collections import defaultdict
from typing import TypedDict, Callable, Self

try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from aiohttp import web
from aiohttp.web import middleware
from aiohttp.web_urldispatcher import SystemRoute
from multidict import CIMultiDict

from .consts import EVENT_HANDLER_EVENT_TYPE_MAP

class Event(TypedDict, total=False):
    # Web event sent via htmx extenstion event-header
    # https://developer.mozilla.org/en-US/docs/Web/Events
    # https://htmx.org/extensions/event-header/
    pass

class Gladius:
    app: web.Application
    callbacks: dict[str, dict[str, tuple['Component', Callable]]] # {g_id: {event_type: [component, func]}}
    update_components: dict[str, list['Component']]

    def __init__(self):
        @middleware
        async def gladius_middleware(request: web.Request, handler: Callable) -> web.Response:
            # print(request, handler)
            req: dict

            if request.method == 'GET':
                req = dict(request.query)
            else:
                req = await request.json()

            response: web.Response

            # default aiohttp request handler
            excs = None

            try:
                response = await handler(request)
                return response
            except Exception as e:
                excs = [e]

            # gladius handler
            try:
                res: str | dict = await handler(self, req)
            except Exception as e:
                excs = [e]

            # if both throw error, 
            if excs:
                e = ExceptionGroup(f'request {request}, handler {handler} error', excs)
                raise e

            if isinstance(res, str):
                response = web.Response(text=res, content_type='text/html')
            elif isinstance(res, dict):
                response = web.json_reponse(res)
            else:
                raise ValueError(res)

            return response

        self.app = web.Application(middlewares=[gladius_middleware])
        
        self.app.add_routes([
            web.static('/static/gladius', os.path.join(os.path.split(__loader__.path)[0], 'static')),
            web.get('/api/1.0/_event/{event_type}/{g_id}', self.get_api_1_0__event),
            web.get('/api/1.0/_event/{event_type}/{g_session_id}/{g_id}', self.get_api_1_0__session_event),
            web.post('/api/1.0/_event/{event_type}/{g_id}', self.post_api_1_0__event),
        ])

        if os.path.exists('static'):
            self.app.add_routes([
                web.static('/static', 'static'),
            ])

        self.callbacks = defaultdict(dict)
        self.update_components = {}

    def add_route(self, method, path, handler):
        routes = [web.route(method.lower(), path, handler)]
        self.app.add_routes(routes)

    def route(self, path, component: 'Component'):
        async def _get_path_handler(request: web.Request) -> web.Response:
            # print('_get_path_handler:', request)
            res: str = component.render()
            response = web.Response(text=res, content_type='text/html')
            return response

        self.app.router.add_get(path, _get_path_handler)

    async def get_api_1_0__event(self, request: web.Request) -> web.Response:
        event_type: str = request.match_info['event_type']
        g_id: str = request.match_info['g_id']
        g_session_id: str = request.headers['G-Session-ID']
        # print('get_api_1_0__event:', request, event_type, g_id, g_session_id)
        
        component, callback = self.callbacks[g_id][event_type]
        await callback(component, {})
        
        res: str = component.render()
        # print('get_api_1_0__event:', res)
        response = web.Response(text=res, content_type='text/html')
        return response

    async def get_api_1_0__session_event(self, request: web.Request) -> web.Response:
        event_type: str = request.match_info['event_type']
        g_id: str = request.match_info['g_id']
        g_session_id: str = request.match_info['g_session_id']
        # print('get_api_1_0__session_event:', request, event_type, g_id, g_session_id)
        
        component, callback = self.callbacks[g_id][event_type]
        await callback(component, {})
        
        res: str = component.render()
        # print('get_api_1_0__session_event:', res)
        response = web.Response(text=res, content_type='text/html')
        return response

    async def post_api_1_0__event(self, request: web.Request) -> web.Response:
        event_type: str = request.match_info['event_type']
        g_id: str = request.match_info['g_id']
        g_session_id: str = request.headers['G-Session-ID']
        event: Event = request.headers['Triggering-Event']
        # print('post_api_1_0__event:', request, event_type, g_id, g_session_id, event)

        # self.update_components[g_session_id] = []
        component, callback = self.callbacks[g_id][event_type]
        res: dict = await callback(component, event)

        # print('post_api_1_0__event:', res)
        return web.json_response(res)

    def get_app(self):
        return self.app

    def run_app(self, *args, **kwargs):
        web.run_app(self.app, *args, **kwargs)



__all__ = [
    'Page',
    'Navbar',
    'NavbarButton',
    'Link',
    'Flex',
    'VFlex',
    'Card',
    'Join',
    'VJoin',
    'Button',
    'Table',
    'Text',
    'Artboard',
    'Divider',
    'Footer',
    'FooterTitle',
    'Hero',
    'HeroContent',
    'Indicator',
    'IndicatorItem',
    'Stack',
    'Toast',
    'Alert',
    'DaisyUI',
]

from typing import Self

from .gladius import Gladius, Event
from .component import Component, ComponentLibrary
from . import html5

class Page(html5.Page):
    # NOTE: high-level component
    default_class: str = 'container mx-auto'

    def __init__(self, component_library: ComponentLibrary, **kwargs):
        super().__init__(component_library, **kwargs)
        h: ComponentLibrary = html5.Html5(component_library.ctx)
        self.head.add(link := h.Link(href='https://cdn.jsdelivr.net/npm/daisyui@3.8.0/dist/full.css', rel='stylesheet', type='text/css'))
        self.head.add(script := h.Script(src='https://cdn.tailwindcss.com'))

class Navbar(html5.Div):
    default_class: str = 'navbar bg-base-100'

    def add(self, child: Component) -> Self:
        child.set_attr(hx_target='body')
        self.children.append(child)
        return self

class NavbarButton(html5.A):
    # NOTE: extended component
    default_class: str = 'btn btn-ghost normal-case text-xl'

class Link(html5.A):
    default_class: str = 'link'

class Flex(html5.Div):
    default_class: str = 'flex w-full gap-1'

class VFlex(Flex):
    # NOTE: extended component
    default_class: str = 'flex flex-col w-full gap-1'

class Card(html5.Div):
    default_class: str = 'card grid h-20 flex-grow bg-base-300 rounded-box place-items-center'

class Join(html5.Div):
    default_class: str = 'join'

    def add(self, child: Component) -> Self:
        if not child.has_class('join-item'):
            child.add_class('join-item')
        
        child.set_attr(hx_target='body')
        self.children.append(child)
        return self

class VJoin(Join):
    # NOTE: extended component
    default_class: str = 'join join-vertical'

class Button(html5.Button):
    default_class: str = 'btn'

class Table(Component):
    header: list[str]
    rows: list[list]

    def __init__(self, component_library: ComponentLibrary, header: list[str]=[], rows: list[list]=[], **kwargs):
        super().__init__(component_library, **kwargs)
        self.header = header
        self.rows = rows

    def render(self) -> str:
        rendered_header = '\n'.join([
            '<tr>',
            '\n'.join(f'<th>{h}</th>' for h in self.header),
            '</tr>',
        ])
        
        rendered_rows = '\n'.join([
            '\n'.join([
                '<tr>',
                '\n'.join(f'<td>{v}</td>' for v in row),
                '</tr>',
            ]) for row in self.rows
        ])

        return f'''
            <table class="table" {self.render_attrs()}>
                <thead class="bg-base-200">
                    {rendered_header}
                </thead>
                <tbody> 
                    {rendered_rows}
                </tbody>
            </table>
        '''

class Text(html5.Span):
    # NOTE: extended component
    pass

#
# Layout
#
class Artboard(html5.Div):
    default_class: str = 'artboard'

class Divider(html5.Div):
    default_class: str = 'divider'

# TODO: Drawer

class Footer(html5.Footer):
    default_class: str = 'footer'

class FooterTitle(html5.Header):
    # NOTE: extended component
    default_class: str = 'footer-title'

class Hero(html5.Div):
    default_class: str = 'hero bg-base-200'

class HeroContent(html5.Div):
    default_class: str = 'hero-content'

class Indicator(html5.Div):
    default_class: str = 'indicator'

class IndicatorItem(html5.Span):
    default_class: str = 'indicator-item'

# TODO: Mask

class Stack(html5.Div):
    default_class: str = 'stack'

class Toast(html5.Div):
    default_class: str = 'toast'

class Alert(html5.Div):
    default_class: str = 'alert'

#
# DaisyUI Component Library
#
class DaisyUI(ComponentLibrary):
    def __init__(self, ctx: Gladius):
        super().__init__(ctx)

        self.component_map: dict[str, Component] = {
            k: v
            for k, v in dict(globals()).items()
            if isinstance(v, type) and issubclass(v, Component)
        }

    def __getattr__(self, attr):
        ComponentType: type = self.get_component_type(attr)

        def _component(*args, **kwargs) -> Component:
            return ComponentType(self, *args, **kwargs)

        return _component



__all__ = [
    'Page',
    'Html',
    'Head',
    'Meta',
    'Link',
    'Title',
    'Script',
    'Body',
    'Div',
    'Span',
    'Footer',
    'Nav',
    'Header',
    'H1',
    'H2',
    'H3',
    'H4',
    'P',
    'A',
    'BlockQuote',
    'Figure',
    'FigCaption',
    'Strong',
    'Em',
    'Code',
    'Pre',
    'Ol',
    'Ul',
    'Li',
    'Table',
    'THead',
    'Tr',
    'Th',
    'Td',
    'Img',
    'Video',
    'Source',
    'Hr',
    'Button',
    'Main',
    'HGroup',
    'Abbr',
    'B',
    'I',
    'Cite',
    'Del',
    'Ins',
    'Kbd',
    'Mark',
    'S',
    'Small',
    'Sub',
    'Sup',
    'U',
    'Input',
    'Html5',
]

from uuid import uuid4
from typing import Self

from .gladius import Gladius, Event
from .component import Component, ComponentLibrary

class Page(Component):
    # NOTE: high-level component
    tag: str = 'html'
    title: str
    favicon: str
    html: 'Html'
    head: 'Head'
    body: 'Body'

    def __init__(self, component_library: 'ComponentLibrary', title: str='Gladius', favicon: str='/static/gladius/favicon.png', **kwargs):
        super().__init__(component_library, **kwargs)
        self.title = title
        self.favicon = favicon
        h: ComponentLibrary = Html5(component_library.ctx)

        # html
        html = h.Html()
        self.html = html
        
        # head
        html.add(head := h.Head())
        head.add(meta := h.Meta(charset='utf-8'))
        head.add(meta := h.Meta(name='viewport', content_='width=device-width'))
        head.add(title := h.Title(content=self.title))
        head.add(link := h.Link(rel='shortcut icon', type='image/png', href=self.favicon))
        head.add(script := h.Script(src='https://unpkg.com/htmx.org@1.9.6'))
        head.add(script := h.Script(src='https://unpkg.com/htmx.org/dist/ext/debug.js'))
        head.add(script := h.Script(src='https://unpkg.com/htmx.org/dist/ext/json-enc.js'))
        head.add(script := h.Script(src='https://unpkg.com/htmx.org/dist/ext/event-header.js'))
        head.add(script := h.Script(src='https://unpkg.com/idiomorph/dist/idiomorph-ext.min.js'))
        head.add(script := h.Script(src='/static/gladius/multi-path-deps.js'))
        self.head = head

        # body
        html.add(body := h.Body(hx_ext='multi-path-deps,morph', hx_boost='true', hx_swap='morph:innerHTML'))
        self.body = body

    def add(self, child: 'Component') -> Self:
        self.body.children.append(child)
        return self

    def render(self) -> str:
        # create new g_session_id for each GET request
        g_session_id = str(uuid4())
        self.html.set_attr(g_session_id=g_session_id)
        return self.html.render()

class Html(Component):
    tag: str = 'html'
    default_attrs: dict = {'lang': 'en-US'}

    def render(self) -> str:
        return f'''
            <!doctype html>
            <{self.tag} {self.render_attrs()}>
                {self.render_children()}
            </{self.tag}>
        '''

class Head(Component):
    tag: str = 'head'

class Meta(Component):
    tag: str = 'meta'
    void_element: bool = True

class Link(Component):
    tag: str = 'link'
    void_element: bool = True

class Title(Component):
    tag: str = 'title'

class Script(Component):
    tag: str = 'script'

class Body(Component):
    tag: str = 'body'

    def __init__(self, component_library: 'ComponentLibrary', **kwargs):
        super().__init__(component_library, **kwargs)

        # content change
        async def _oncontentchange(content: str, event: Event):
            # print('_oncontentchange', content, event)
            pass

        event_type: str = '_oncontentchange'
        self.attrs[event_type] = _oncontentchange

class Div(Component):
    tag: str = 'div'

class Span(Component):
    tag: str = 'span'

class Footer(Component):
    tag: str = 'footer'

class Nav(Component):
    tag: str = 'nav'

class Header(Component):
    tag: str = 'header'

class H1(Component):
    tag: str = 'h1'

class H2(Component):
    tag: str = 'h2'

class H3(Component):
    tag: str = 'h3'

class H4(Component):
    tag: str = 'h4'

class H5(Component):
    tag: str = 'h5'

class H6(Component):
    tag: str = 'h6'

class P(Component):
    tag: str = 'p'

class A(Component):
    tag: str = 'a'

class BlockQuote(Component):
    tag: str = 'blockquote'

class Figure(Component):
    tag: str = 'figure'

class FigCaption(Component):
    tag: str = 'figcaption'

class Strong(Component):
    tag: str = 'strong'

class Em(Component):
    tag: str = 'em'

class Code(Component):
    tag: str = 'code'

class Pre(Component):
    tag: str = 'pre'

class Ol(Component):
    tag: str = 'ol'

class Ul(Component):
    tag: str = 'ul'

class Li(Component):
    tag: str = 'li'

class Table(Component):
    tag: str = 'table'

class THead(Component):
    tag: str = 'thead'

class Tr(Component):
    tag: str = 'tr'

class Th(Component):
    tag: str = 'th'

class Td(Component):
    tag: str = 'td'

class Img(Component):
    tag: str = 'img'
    void_element: bool = True

class Video(Component):
    tag: str = 'video'

class Source(Component):
    tag: str = 'source'
    void_element: bool = True

class Hr(Component):
    tag: str = 'hr'
    void_element: bool = True

class Button(Component):
    tag: str = 'button'

class Main(Component):
    tag: str = 'main'

class HGroup(Component):
    tag: str = 'hgroup'

class Abbr(Component):
    tag: str = 'abbr'

class B(Component):
    tag: str = 'b'

class I(Component):
    tag: str = 'I'

class Cite(Component):
    tag: str = 'cite'

class Del(Component):
    tag: str = 'del'

class Ins(Component):
    tag: str = 'ins'

class Kbd(Component):
    tag: str = 'kbd'

class Mark(Component):
    tag: str = 'mark'

class S(Component):
    tag: str = 's'

class Small(Component):
    tag: str = 'small'

class Sub(Component):
    tag: str = 'sub'

class Sup(Component):
    tag: str = 'sup'

class U(Component):
    tag: str = 'u'

class Input(Component):
    tag: str = 'input'
    void_element: bool = True

#
# Html5 Component Library
#
class Html5(ComponentLibrary):
    def __init__(self, ctx: Gladius):
        super().__init__(ctx)

        self.component_map: dict[str, Component] = {
            k: v
            for k, v in dict(globals()).items()
            if isinstance(v, type) and issubclass(v, Component)
        }

    def __getattr__(self, attr):
        ComponentType: type = self.get_component_type(attr)

        def _component(*args, **kwargs) -> Component:
            return ComponentType(self, *args, **kwargs)

        return _component



__all__ = [
    'Page',
    'Main',
    'Grid',
    'Headings',
    'Pico',
]

from typing import Self

from .gladius import Gladius, Event
from .component import Component, ComponentLibrary
from . import html5

class Page(html5.Page):
    # NOTE: high-level component
    default_class: str = 'container mx-auto'

    def __init__(self, component_library: ComponentLibrary, **kwargs):
        super().__init__(component_library, **kwargs)
        h: ComponentLibrary = html5.Html5(component_library.ctx)
        self.head.add(link := h.Link(href='https://cdn.jsdelivr.net/npm/@picocss/pico@1/css/pico.min.css', rel='stylesheet', type='text/css'))

#
# layout
#
class Main(html5.Main):
    default_class: str = 'container'

class Grid(html5.Div):
    default_class: str = 'grid'

class Div(html5.Div):
    pass

#
# typography
#
class H1(html5.H1): pass
class H2(html5.H2): pass
class H3(html5.H3): pass
class H4(html5.H4): pass
class H5(html5.H5): pass
class H6(html5.H6): pass
class HGroup(html5.HGroup): pass

class Headings(html5.Div):
    default_class: str = 'headings'

class P(html5.P): pass
class Abbr(html5.Abbr): pass
class Strong(html5.Strong): pass
class B(html5.B): pass
class I(html5.I): pass
class Em(html5.Em): pass
class Cite(html5.Cite): pass
class Del(html5.Del): pass
class Ins(html5.Ins): pass
class Kbd(html5.Kbd): pass
class Mark(html5.Mark): pass
class S(html5.S): pass
class Small(html5.Small): pass
class Sub(html5.Sub): pass
class Sup(html5.Sup): pass
class U(html5.U): pass
class Link(html5.A): pass

class SecondaryLink(Link):
    default_class: str = 'secondary'

class ContrastLink(Link):
    default_class: str = 'contrast'

class BlockQuote(html5.BlockQuote): pass
class Footer(html5.Footer): pass

#
# buttons
#
class Button(html5.Button): pass

class SubmitInput(html5.Input):
    def __init__(self, component_library: ComponentLibrary, **kwargs):
        super().__init__(component_library, type='submit', **kwargs)

class ButtonLink(html5.A):
    def __init__(self, component_library: ComponentLibrary, content: str='', **kwargs):
        super().__init__(component_library, role='button', content=content, **kwargs)

class SecondaryButtonLink(ButtonLink):
    default_class: str = 'secondary'

class ContrastButtonLink(ButtonLink):
    default_class: str = 'contrast'

class OutlineButtonLink(ButtonLink):
    default_class: str = 'outline'

class SecondaryOutlineButtonLink(ButtonLink):
    default_class: str = 'secondary outline'

class ContrastOutlineButtonLink(ButtonLink):
    default_class: str = 'contrast outline'

#
# Pico Component Library
#
class Pico(ComponentLibrary):
    def __init__(self, ctx: Gladius):
        super().__init__(ctx)

        self.component_map: dict[str, Component] = {
            k: v
            for k, v in dict(globals()).items()
            if isinstance(v, type) and issubclass(v, Component)
        }

    def __getattr__(self, attr):
        ComponentType: type = self.get_component_type(attr)

        def _component(*args, **kwargs) -> Component:
            return ComponentType(self, *args, **kwargs)

        return _component



__all__ = [
    'Component',
    'ComponentLibrary',
]

import json
from uuid import uuid4
from copy import copy, deepcopy
from typing import Union, Self, Any

from .consts import EVENT_HANDLER_EVENT_TYPE_MAP
from .gladius import Gladius, Event

class Component:
    component_library: 'ComponentLibrary'
    attrs: dict
    content: str = ''
    children: list['Component']
    tag: str = 'div'
    default_class: str = ''
    default_attrs: dict[str, str] = {}

    # A void element is an element in HTML that cannot have any child nodes
    # (i.e., nested elements or text nodes).
    # Void elements only have a start tag;
    # end tags must not be specified for void elements.
    # https://developer.mozilla.org/en-US/docs/Glossary/Void_element
    void_element: bool = False
    
    def __init__(self, component_library: 'ComponentLibrary', content: str='', **kwargs):
        self.component_library = component_library
        
        # attrs
        attrs = {**kwargs}

        if 'class' not in attrs and 'class_' not in attrs:
            attrs['class'] = self.default_class

        attrs['g-id'] = str(uuid4())
        self.attrs = {}
        self.set_attr(**attrs)

        # children
        self.content = content
        self.children = []

    def set_attr(self, **kwargs) -> Self:
        # replace _ with -
        attrs = {
            k.replace('_', '-'): v
            for k, v in kwargs.items()
            if k not in ('class', 'class_', 'id', 'id_')
        }

        # class
        if 'class' in kwargs:
            attrs['class'] = kwargs['class']
        elif 'class_' in kwargs:
            attrs['class'] = kwargs['class_']

        # id
        if 'id' in kwargs:
            attrs['id'] = kwargs['id']
        elif 'id_' in kwargs:
            attrs['id'] = kwargs['id_']

        self.attrs.update(attrs)
        return self

    def get_attr(self, attr: str, default_value: str|None=None) -> Any:
        return self.attrs.get(attr, default_value)

    def del_attr(self, attr: str):
        del self.attrs[attr]

    def has_attr(self, attr: str) -> bool:
        return attr in self.attrs

    def add_class(self, class_: str) -> Self:
        component_class = self.attrs.get('class', '')
        class_ = component_class + ' ' + class_
        class_ = class_.strip()
        self.attrs['class'] = class_
        return self

    def remove_class(self, class_: str) -> Self:
        class_ = class_.strip()
        component_class = self.attrs.get('class', '')
        component_classes = [n for n in component_class.split() if n != class_]
        class_ = ' '.join(component_classes)
        self.attrs['class'] = class_
        return self

    def has_class(self, class_: str) -> bool:
        class_ = class_.strip()
        return class_ in self.attrs.get('class', '')

    def add(self, child: 'Component') -> Self:
        self.children.append(child)
        return self

    def remove(self, child: 'Component') -> Self:
        self.children.remove(child)
        return self

    def clone(self, deep: bool=True) -> Self:
        return deepcopy(self) if deep else copy(self)

    def _render_value(self, k, v) -> str:
        r: str

        if isinstance(v, str):
            r = json.dumps(v)
        else:
            r = f"'{json.dumps(v)}'"

        return r

    def _render_attr(self, k, v):
        attr: str
        g_id: str = self.attrs['g-id']

        if not k.startswith('_') and k in EVENT_HANDLER_EVENT_TYPE_MAP and callable(v):
            # standard events
            event_type: str = EVENT_HANDLER_EVENT_TYPE_MAP[k]

            attr = ' '.join([
                f'hx-trigger="{event_type}"',
                f'hx-post="/api/1.0/_event/{event_type}/{g_id}"',
                'hx-ext="json-enc,event-header"',
                'hx-swap="none"',
                'hx-headers=\'js:{"G-Session-ID": document.body.getAttribute("g-session-id")}\'',
            ])

            self.component_library.ctx.callbacks[g_id][event_type] = [self, v]
        elif k.startswith('_') and k in EVENT_HANDLER_EVENT_TYPE_MAP and callable(v):
            # custom events
            event_type: str = EVENT_HANDLER_EVENT_TYPE_MAP[k]

            attr = ' '.join([
                'hx-trigger="multi-path-deps"',
                f'hx-get="/api/1.0/_event/{event_type}/{g_id}"',
                'multi-path-deps=\'["/api/1.0/_event"]\'',
                f'hx-target=\'[g-id="{g_id}"]\'',
                'hx-swap="outerHTML"',
                'hx-headers=\'js:{"G-Session-ID": document.documentElement.getAttribute("g-session-id")}\'',
            ])

            self.component_library.ctx.callbacks[g_id][event_type] = [self, v]
        else:
            # other attrs - non-event/non-callback attrs
            attr = f'{k}={self._render_value(k, v)}'
        
        return attr

    def render_attrs(self) -> str:
        rendered_attrs: str = ' '.join(self._render_attr(k, v) for k, v in self.attrs.items())
        return rendered_attrs

    def render_children(self) -> str:
        rendered_children: str = '\n'.join(c.render() for c in self.children)
        return rendered_children

    def render(self) -> str:
        if self.void_element:
            return f'''
                <{self.tag} {self.render_attrs()} />
            '''
        else:
            return f'''
                <{self.tag} {self.render_attrs()}>
                    {self.content}
                    {self.render_children()}
                </{self.tag}>
            '''

class ComponentLibrary:
    ctx: Gladius
    component_map: dict[str, type]

    def __init__(self, ctx: Gladius):
        self.ctx = ctx

        component_map: dict[str, Component] = {
            k: v
            for k, v in dict(globals()).items()
            if isinstance(v, type) and issubclass(v, Component)
        }

        self.component_map = component_map

    def add_component_type(self, component_name: str, component_type: type):
        self.component_map[component_name] = component_type

    def get_component_type(self, component_name: str) -> type:
        return self.component_map[component_name]
